/*
 * Sample program to show the use of liboctrope.a
 *
 * $Id: ropelength.c,v 1.7 2005/03/02 19:34:45 ashted Exp $
 *
 */

/* Copyright 2004 The University of Georgia. */

/* This file is part of liboctrope.
   
liboctrope is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

liboctrope is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with liboctrope; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#ifdef HAVE_FLOAT_H
#include <float.h>
#endif
#include "octrope.h"
#include <argtable2.h>

int main(int argc,char *argv[]) {

  octrope_link *L;
  FILE *infile;
  double curve_length;
  double thickness;
  double ropelength;
  double min_rad;
  double min_strut;

  struct arg_file *knotfiles = arg_filen(NULL,NULL,"<file>",1,100,
    "input file (up to 100)");
  struct arg_int *d = arg_int0("dD",NULL,"<int>","debug level (>=0)");
  struct arg_lit *dbg = arg_lit0(NULL,"debug","same as -d 4");
  struct arg_int *l = arg_int0("l","levels","<int>","levels in octree (>=0)");
  struct arg_dbl *lam = arg_dbl0("L","lambda","<dbl>","lambda (default=1)");
  struct arg_lit *help = arg_lit0("h","help","print this help and exit");
  struct arg_end *end = arg_end(20);

  void *argtable[] = {lam,l,d,dbg,help,knotfiles,end};
  int nerrors;
  int filecnt;

  if (arg_nullcheck(argtable) != 0) {
    /* NULL entries detected, allocations must have failed. */
    fprintf(stderr,"%s: insufficient memory\n",argv[0]);
    arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));
    return 1;
  }
 
  /* Set default lambda */
  lam->dval[0] = 1;
  
  /* Parse the command line as defined by argtable[] */
  nerrors = arg_parse(argc,argv,argtable);

  /* special case: '--help' takes preceence over error reporting */
  if (help->count > 0) {
    fprintf(stderr,"Usage: %s ",argv[0]);
    arg_print_syntax(stderr,argtable,"\n");
    arg_print_glossary(stderr,argtable,"  %-25s %s\n");
    arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));
    return 0;
  }

  /* If the parser returned any errors then display them and exit */
  if (nerrors > 0) {
    /* Display the error details contained in the arg_end struct.*/
    fprintf(stderr,"\n");
    arg_print_errors(stderr,end,argv[0]);
    fprintf(stderr,"\nUsage: %s ",argv[0]);
    arg_print_syntax(stderr,argtable,"\n");
    arg_print_glossary(stderr,argtable,"  %-25s %s\n");
    arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));
    return 1;
  }

  if (dbg->count > 0) {
    fprintf(stderr,"%s:Setting debug level to 4.\n",argv[0]);
    octrope_set_debug(4);
  } 
  if (d->count > 0) {
    fprintf(stderr,"%s:Setting debug level to %d.\n",argv[0],
      d->ival[d->count-1]);
    octrope_set_debug(d->ival[d->count-1]);
  }
  if (l->count > 0) {
    if (octrope_debug_level() > 2) {
      fprintf(stderr,"%s:Setting levels to %d\n",argv[0],
        l->ival[l->count-1]);
    }
    octrope_set_levels(l->ival[l->count-1]); 
  }

  for (filecnt = 0; filecnt < knotfiles->count; filecnt++) {
    infile = fopen(knotfiles->filename[filecnt],"r");
    if (infile == NULL) {
      fprintf(stderr,"%s: Couldn't find input file %s.\n",argv[0],
        knotfiles->filename[filecnt]);
      return 1;
    }

    L = octrope_link_read(infile);
    fclose(infile);
    if (octrope_error_num != 0) { 
      fprintf(stderr,"%s:%s",argv[0],octrope_error_str); 
      arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));
      return 1;
    }

    curve_length = octrope_curvelength(L);
    if (octrope_error_num != 0) {
      fprintf(stderr,"%s:%s",argv[0],octrope_error_str); 
      octrope_link_free(L);
      if (octrope_error_num != 0) {
        fprintf(stderr,"%s:%s",argv[0],octrope_error_str); 
      }
      arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));
      return 1;
    }
    thickness = octrope_thickness(L,lam->dval[0],NULL,0);
    if (octrope_error_num != 0) {
      fprintf(stderr,"%s:%s",argv[0],octrope_error_str); 
      octrope_link_free(L);
      if (octrope_error_num != 0) {
        fprintf(stderr,"%s:%s",argv[0],octrope_error_str); 
      }
      arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));
      return 1;
    }
    if (thickness == 0) {
      printf("%s:Ropelength: infinity  Thickness: %f\n",
        knotfiles->filename[filecnt], thickness);
    } else {
      ropelength = curve_length/thickness;
      printf("%s:Ropelength: %f  Thickness: %f\n",
        knotfiles->filename[filecnt], ropelength, thickness);
    }
    /* And now a little extra information, gratis */
    octrope(L,lam->dval[0],NULL,NULL,NULL,&min_rad,&min_strut,0,0,NULL,
            0,NULL,0,0,NULL,0,NULL,NULL,0);
    printf("%s:minRad: %f    minStrut: %f\n",
      knotfiles->filename[filecnt], min_rad, min_strut);

    octrope_link_free(L);
    if (octrope_error_num != 0) {
      fprintf(stderr,"%s:%s",argv[0],octrope_error_str); 
      return 1;
    }
  }

  return 0;
}
