/* 
 * Take the given link, replacing each corner with its minrad circle arc (not
 * using the global MinRad value).  Place enough points along the resulting
 * curve(s) so that using an n^2 algorithm to check distances, we can find
 * its ropelength to within the desired accuracy.
 *
 * $Id$
 *
 */

/* Copyright 2004 The University of Georgia. */

/* This file is part of liboctrope.
   
liboctrope is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

liboctrope is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with liboctrope; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#define PI 3.141592653589793238462643383279502884197169399375

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#ifdef HAVE_FLOAT_H
#include <float.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include "octrope.h"
#include <argtable2.h>
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif

typedef struct point_n_curvature_type {
  octrope_vector point;
  double curvature;
} point_n_curvature;

void add_arc_points(point_n_curvature *points, int *num_points,
                              octrope_pline P, int vert, double accuracy,
                              double *total_len, double *total_curvature) {
  /* Add enough points from this arc to keep us within accuracy */
  double normin,normout;
  octrope_vector in,out,cross_prod,half_side,center, first_rad, second_rad;
  double dot_prod,cross_prod_norm,rad;
  double radius, scale;
  double alpha, arc_len, angle;
  int steps,cnt;
  double halflength, position;
  octrope_vector half_shortside;

  in = octrope_vminus(P.vt[vert],P.vt[vert-1]);
  out = octrope_vminus(P.vt[vert+1],P.vt[vert]);
  normin = octrope_norm(in);
  normout = octrope_norm(out);
  
  dot_prod = octrope_dot(in,out);
  cross_prod = octrope_cross(in,out);
  cross_prod_norm = octrope_norm(cross_prod);
  if (fabs(cross_prod_norm) < 1E-12) {
    /* This must be a straight line */
    halflength = (normin < normout) ? normin/2 : normout/2;
    for (position = normin - halflength; position < normin; position += accuracy) {
      octrope_vweighted(points[*num_points].point,position/normin,
                        P.vt[vert-1],P.vt[vert]);
      points[*num_points].curvature = *total_curvature;
      (*num_points)++;
    }
    for (position = 0; position < halflength; position += accuracy) {
      octrope_vweighted(points[*num_points].point,position/normout,
                        P.vt[vert],P.vt[vert+1]);
      points[*num_points].curvature = *total_curvature;
      (*num_points)++;
    }
    return;
  }
  /* Angle subtended by the minrad arc */
  alpha = atan2(cross_prod_norm,dot_prod); 
  rad = (normin*normout + dot_prod)/(2*cross_prod_norm);

  radius = rad*((normin < normout) ? normin : normout);
  first_rad = octrope_cross(in,cross_prod);
  scale = radius/octrope_norm(first_rad); /* For the macro's sake */
  octrope_vsmult(scale,first_rad);
  second_rad = octrope_cross(out,cross_prod);
  scale = radius/octrope_norm(second_rad); /* For the macro's sake */
  octrope_vsmult(scale,second_rad);
  if (normin < normout) {
    center = octrope_scalarmult(-1,first_rad);
    octrope_vmadd(center,1/2,in);
    octrope_vweighted(half_shortside,0.5,P.vt[vert-1],P.vt[vert]);
    octrope_vadd(center,half_shortside);
  } else {
    center = octrope_scalarmult(-1,second_rad);
    octrope_vmadd(center,1/2,out);
    octrope_vweighted(half_shortside,0.5,P.vt[vert],P.vt[vert+1]);
    octrope_vadd(center,half_shortside);
  }
  arc_len = radius*alpha; /* Length of minrad arc */
  *total_len += arc_len;
  steps = 2*arc_len/accuracy;
  for (cnt = 0; cnt < steps; cnt++) {
    angle = cnt*alpha/steps;
    points[*num_points].point = center;
    octrope_vmadd(points[*num_points].point,cos(angle),first_rad);
    octrope_vmadd(points[*num_points].point,sin(angle)*radius/normin,in);
    points[*num_points].curvature = *total_curvature+angle;
    (*num_points)++;
  }
  *total_curvature += alpha;
}

int main(int argc,char *argv[]) {

  octrope_link *L;
  FILE *infile;
  FILE *outfile;

  struct arg_file *knotfiles = 
    arg_filen(NULL,NULL,"<file>",1,1024,"input .VECT file");
  struct arg_file *outputfile =
    arg_file0("o","output","<file>","output .VECT file (for debugging)");
  struct arg_dbl *acc = arg_dbl0("a","accuracy","<dbl>",
    "Tolerance for length calculation (default=1e-6");
  struct arg_lit *help = arg_lit0("h","help","print this help and exit");
  struct arg_end *end = arg_end(20);

  void *argtable[] = {help,acc,outputfile,knotfiles,end};
  int nerrors;
  int filecnt;
  int *np_array;
  double *curvatures;

  point_n_curvature **points;
  int num_points, comp, vert, cnt, scale_factor;
  int start_vert, end_vert;
  double total_len, total_curvature;
  double inlen, outlen, position;
  octrope_vector pos_vect;
  double side_len, longest_side;
  int point1, point2;
  double shortest_distance, dist;

  if (arg_nullcheck(argtable) != 0) {
    /* NULL entries detected, allocations must have failed. */
    fprintf(stderr,"%s: insufficient memory\n",argv[0]);
    arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));
    return 1;
  }

  /* Set defaults */
  acc->dval[0] = 1e-6;
  outputfile->filename[0]="-";

  /* Parse the command line as defined by argtable[] */
  nerrors = arg_parse(argc,argv,argtable);

  /* special case: '--help' takes preceence over error reporting */
  if (help->count > 0) {
    fprintf(stderr,"Usage: %s ",argv[0]);
    arg_print_syntax(stderr,argtable,"\n");
    arg_print_glossary(stderr,argtable,"  %-25s %s\n");
    arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));
    return 0;
  }

  /* If the parser returned any errors then display them and exit */
  if (nerrors > 0) {
    /* Display the error details contained in the arg_end struct.*/
    fprintf(stderr,"\n");
    arg_print_errors(stderr,end,argv[0]);
    fprintf(stderr,"\nUsage: %s ",argv[0]);
    arg_print_syntax(stderr,argtable,"\n");
    arg_print_glossary(stderr,argtable,"  %-25s %s\n");
    arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));
    return 1;
  }

  for (filecnt = 0; filecnt < knotfiles->count; filecnt++) {
    infile = fopen(knotfiles->filename[filecnt],"r");
    if (infile == NULL) {
      fprintf(stderr,"%s: Couldn't find input file %s.\n",argv[0],
        knotfiles->filename[filecnt]);
      return 1;
    }

    L = octrope_link_read(infile);
    fclose(infile);
    if (octrope_error_num != 0) { 
      fprintf(stderr,"%s:%s",argv[0],octrope_error_str); 
      arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));
      return 1;
    }

    /* Find the longest side so as to know how many points we need */

    longest_side = 0;
    num_points = 0;
    if ((points = malloc(L->nc*sizeof(point_n_curvature *))) == NULL) {
      fprintf(stderr,"Error, unable to allocate space for %d lists of points.\n", L->nc);
      exit(1);
    }
    if ((np_array = malloc(L->nc*sizeof(int))) == NULL) {
      fprintf(stderr,"Error, unable to allocate space for %d point counts.\n", L->nc);
      exit(1);
    }
    if ((curvatures = malloc(L->nc*sizeof(double))) == NULL) {
      fprintf(stderr,"Error, unable to allocate space for %d curvatures.\n", L->nc);
      exit(1);
    }
    for (comp = 0; comp < L->nc; comp++) {
      num_points += L->cp[comp].nv;
      if (L->cp[comp].acyclic) {
        start_vert = 1;
        end_vert = L->cp[comp].nv-1;
      } else {
        start_vert = 0;
        end_vert = L->cp[comp].nv;
      }
      for (vert = start_vert; vert < end_vert; vert++) {
        side_len =
          octrope_norm(octrope_vminus(L->cp[comp].vt[vert-1],L->cp[comp].vt[vert]));
        if (side_len > longest_side) {
          longest_side = side_len;
        }
      }
    }
    scale_factor = ceil(longest_side/acc->dval[0])*2;
    if (num_points*scale_factor >= INT32_MAX) {
      fprintf(stderr,"Requested accuracy results in too many points.\n");
      fprintf(stderr,"Use a number greater than %f.\n",num_points*longest_side/INT32_MAX);
      exit(1);
    }
    for (comp = 0; comp < L->nc; comp++) {
      printf("Allocating %d.\n",L->cp[comp].nv*scale_factor);
      if ((points[comp] = 
           malloc(L->cp[comp].nv*scale_factor*sizeof(point_n_curvature))) == NULL) {
        fprintf(stderr,"Error, unable to allocate space for %d vectors.\n",
          L->cp[comp].nv*scale_factor);
        fprintf(stderr,"You may wish to reduce the accuracy.\n");
        exit(1);
      }
      np_array[comp] = 0;
    }

    printf("Creating point list.\n");
    num_points = 0;
    total_len = 0;
    for (comp = 0; comp < L->nc; comp++) {
      total_curvature = 0;
      if (L->cp[comp].acyclic) {
        start_vert = 1;
        end_vert = L->cp[comp].nv-1;
      } else {
        start_vert = 0;
        end_vert = L->cp[comp].nv;
      }
      for (vert = start_vert; vert < end_vert; vert++) {
        inlen = octrope_norm(octrope_vminus(L->cp[comp].vt[vert],L->cp[comp].vt[vert-1]));
        outlen = octrope_norm(octrope_vminus(L->cp[comp].vt[vert+1],L->cp[comp].vt[vert]));
        /* Do the half straight-segment preceeding the curve */
        if (inlen - outlen > 2*acc->dval[0]) {
          for (position = inlen/2; 
               position < inlen - (outlen/2); 
               position += acc->dval[0]) {
            octrope_vweighted(points[comp][np_array[comp]].point,position/inlen,
                              L->cp[comp].vt[vert-1],L->cp[comp].vt[vert]);
            points[comp][np_array[comp]].curvature = total_curvature;
            np_array[comp]++;
          }
          total_len += inlen/2 - outlen/2;
        }
        /* Do the curve */
        add_arc_points(points[comp],&np_array[comp],L->cp[comp],vert,acc->dval[0],
                       &total_len, &total_curvature);
        /* Do the half straight-segment following the curve */
        if (outlen - inlen > 2*acc->dval[0]) {
          for (position = inlen/2; 
               position < outlen/2;
               position += acc->dval[0]) {
            octrope_vweighted(points[comp][np_array[comp]].point,position/outlen,
                              L->cp[comp].vt[vert],L->cp[comp].vt[vert+1]);
            points[comp][np_array[comp]].curvature = total_curvature;
            np_array[comp]++;
          }
          total_len += outlen/2 - inlen/2;
        }
      }
      curvatures[comp] = total_curvature;
      num_points += np_array[comp];
    }

    /* Produce VECT file */
    if (outputfile->filename[0][0] != '-') {
      printf("Opening output file: %s\n",outputfile->filename[0]);
      outfile = fopen(outputfile->filename[0],"w");
      if (outfile == NULL) {
        fprintf(stderr,"%s: Couldn't create output file %s.\n",argv[0],
          outputfile->filename[0]);
      } else {
        printf("Writing output file: %s\n",outputfile->filename[0]);
        fprintf(outfile,"VECT\n");
        fprintf(outfile,"%d %d %d\n",L->nc,num_points,L->nc);
        for (comp = 0; comp < L->nc; comp++) {
          if (L->cp[comp].acyclic) {
            fprintf(outfile,"%d ",np_array[comp]);
          } else {
            fprintf(outfile,"%d ",-np_array[comp]);
          }
        }
        fprintf(outfile,"\n");
        for (comp = 0; comp < L->nc; comp++) {
          fprintf(outfile,"1 ");
        }
        fprintf(outfile,"\n");
        for (comp = 0; comp < L->nc; comp++) {
          for (cnt = 0; cnt < np_array[comp]; cnt++) {
            fprintf(outfile,"%f %f %f\n",points[comp][cnt].point.c[0],
                                points[comp][cnt].point.c[1],
                                points[comp][cnt].point.c[2]);
          }
        }
        for (comp = 0; comp < L->nc; comp++) {
          fprintf(outfile,"0 0 1 1\n");
        } 
        fclose(outfile);
      }
      printf("File written.\n");
    }

    /* Measure distances */
    printf("Doing distance checks (n is %d)\n",num_points);
    shortest_distance = DBL_MAX;
    for (comp = 0; comp < L->nc; comp++) {
      for (point1 = 0; point1 < np_array[comp]; point1++) {
        if (curvatures[comp] >= PI) { /* otherwise, no struts to itself */
          for (point2 = point1 + 1; 
               point2 < np_array[comp] - 1 &&
               points[comp][point2+1].curvature - points[comp][point1].curvature < PI; 
               point2++) {}
          if (point2 < np_array[comp] - 1 ||
              points[comp][point2].curvature - points[comp][point1].curvature >= PI) {
            for (; point2 < np_array[comp] && 
                   points[comp][point2].curvature <= 
                     curvatures[comp]-PI+points[comp][point1].curvature ;
                   point2++) {
              dist = octrope_norm(octrope_vminus(points[comp][point1].point,
                                                 points[comp][point2].point));
              if (dist < shortest_distance) {
                shortest_distance = dist;
              }
            }
          }
        }
        for (cnt = comp + 1; cnt < L->nc; cnt++) {
          for (point2 = 0; point2 < np_array[cnt]; point2++) {
            dist = octrope_norm(octrope_vminus(points[comp][point1].point,
                                               points[cnt][point2].point));
            if (dist < shortest_distance) {
              shortest_distance = dist;
            }
          }
        }
      }
    }
    printf("Shortest pairwise distance found: %f\n",shortest_distance);
    printf("Curve length: %f\n",total_len);
    printf("Ropelength: %f\n",total_len/shortest_distance);
    printf("Radius Ropelength: %f\n",2*total_len/shortest_distance);

    for (comp = 0; comp < L->nc; comp++) {
      free(points[comp]);
    } 
    free(points);
    free(np_array);
    octrope_link_free(L);
  }
}
